<?php
/**
 * A socket factory class
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2015 Hub Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 * @todo		Find an interface for hub helper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class SocketFactory extends ObjectFactory {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates a valid socket resource from given packae data and protocol
	 *
	 * @param	$packageData		Raw package data
	 * @param	$protocolInstance	An instance of a HandleableProtocol class
	 * @return	$socketResource		Socket resource
	 */
	public static function createSocketFromPackageData (array $packageData, HandleableProtocol $protocolInstance) {
		// Get an instance
		$factoryInstance = new SocketFactory();

		// Construct registry key
		$registryKey = 'socket_' . $protocolInstance->getProtocolName() . '_' . $packageData[NetworkPackage::PACKAGE_DATA_RECIPIENT];

		// Debug message
		//*NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('SOCKET-FACTORY[' . __METHOD__ . ':' . __LINE__ . ']: Trying to find a socket with registryKey=' . $registryKey);

		// Is the key there?
		if (Registry::getRegistry()->instanceExists($registryKey)) {
			// Get container instance
			$containerInstance = Registry::getRegistry()->getInstance($registryKey);

			// Get socket back
			$socketResource = $containerInstance->getSocketResource();

			// Debug message
			//*NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('SOCKET-FACTORY[' . __METHOD__ . ':' . __LINE__ . ']: Using socket ' . $socketResource . '(' . gettype($socketResource) . ') from registry.');
		} else {
			// Construct configuration entry for object factory and get it
			$className = FrameworkConfiguration::getSelfInstance()->getConfigEntry($protocolInstance->getProtocolName() . '_connection_helper_class');

			// Debug message
			//*NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('SOCKET-FACTORY[' . __METHOD__ . ':' . __LINE__ . ']: Going to use class ' . $className . ' for creating a socket resource ...');

			// And call the static method
			$socketResource = call_user_func($className . '::createConnectionFromPackageData', $packageData);

			// Debug message
			//*NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('SOCKET-FACTORY[' . __METHOD__ . ':' . __LINE__ . ']: Created socket ' . $socketResource . '(' . gettype($socketResource) . ') from class ' . $className . '.');

			// Construct container class, this won't be reached if an exception is thrown
			$containerInstance = self::CreateObjectByConfiguredName('socket_container_class', array($socketResource, NULL, $packageData));

			// Register it with the registry
			Registry::getRegistry()->addInstance($registryKey, $containerInstance);

			// Debug message
			//*NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('SOCKET-FACTORY[' . __METHOD__ . ':' . __LINE__ . ']: Socket is now registered in registry.');
		}

		// Return the resource
		return $socketResource;
	}
}

// [EOF]
?>
