<?php
/**
 * A hub-node class for the 'Boot' mode
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2015 Hub Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class HubBootNode extends BaseHubNode implements NodeHelper, Registerable {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this hub-node class
	 *
	 * @param	$requestInstance	An instance of a Requestable class
	 * @return	$nodeInstance		An instance of this hub-node class
	 */
	public static final function createHubBootNode (Requestable $requestInstance) {
		// Get a new instance
		$nodeInstance = new HubBootNode();

		// Set the request instance
		$nodeInstance->setRequestInstance($requestInstance);

		// Return the instance
		return $nodeInstance;
	}

	/**
	 * Method to "bootstrap" the node. This step does also apply provided
	 * command-line arguments stored in the request instance. The regular node
	 * should communicate with the bootstrap-nodes at this point.
	 *
	 * @return	void
	 * @todo	add some more special bootstrap things for this boot node
	 */
	public function doBootstrapping () {
		// Get UNL
		$unl = $this->detectOwnUniversalNodeLocator();

		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('[' . __METHOD__ . ':' . __LINE__ . ']: unl=' . $unl);

		// Now check if the IP address matches one of the bootstrap nodes
		if ($this->ifAddressMatchesBootstrapNodes($unl)) {
			// Get our port from configuration
			$ourPort = $this->getConfigInstance()->getConfigEntry('node_listen_port');

			// Extract port
			$bootPort = substr($this->getBootUniversalNodeLocator(), -strlen($ourPort), strlen($ourPort));

			// Is the port the same?
			if ($bootPort == $ourPort) {
				// It is the same!
				self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: UNL matches bootstrap node ' . $this->getBootUniversalNodeLocator() . '.');

				// Now, does the mode match
				if ($this->getRequestInstance()->getRequestElement('mode') == self::NODE_TYPE_BOOT) {
					// Output debug message
					self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: Our node is a valid bootstrap node.');
				} else {
					// Output warning
					self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: WARNING: Mismatching mode ' . $this->getRequestInstance()->getRequestElement('mode') . '!=' . BaseHubNode::NODE_TYPE_BOOT . ' detected.');
				}
			} else {
				// IP does match, but no port
				self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: WARNING: Our UNL ' . $unl . ' does match a known bootstrap-node but not the port ' . $ourPort . '/' . $bootPort . '.');
			}
		} else {
			// Node does not match any know bootstrap-node
			self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: WARNING: Our UNL ' . $unl . ' does not match any known bootstrap-nodes.');
		}

		// Enable acceptance of announcements
		$this->enableAcceptingAnnouncements();

		// This might not be all...
		$this->partialStub('Please implement more bootsrapping steps.');
	}

	/**
	 * Add some node-specific filters
	 *
	 * @return	void
	 */
	public function addExtraNodeFilters () {
		// Get the application instance from registry
		$applicationInstance = Registry::getRegistry()->getInstance('app');

		// Get the controller from the application
		$controllerInstance = $applicationInstance->getControllerInstance();

		// @TODO Add some filters here
		$this->partialStub('Add some filters here.');
	}

	/**
	 * Adds extra tasks to the given handler for this node
	 *
	 * @param	$handlerInstance	An instance of a HandleableTask class
	 * @return	void
	 */
	public function addExtraTasks (HandleableTask $handlerInstance) {
		// Prepare a task for booting the DHT
		$taskInstance = ObjectFactory::createObjectByConfiguredName('node_dht_bootstrap_task_class');

		// Register it
		$handlerInstance->registerTask('dht_bootstrap', $taskInstance);
	}
}

// [EOF]
?>
