<?php
/**
 * A general hub node class
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2015 Hub Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
class BaseHubNode extends BaseHubSystem implements Updateable, AddableCriteria {
	/**
	 * Node types
	 */
	const NODE_TYPE_BOOT    = 'boot';
	const NODE_TYPE_MASTER  = 'master';
	const NODE_TYPE_LIST    = 'list';
	const NODE_TYPE_REGULAR = 'regular';

	// Exception constants
	const EXCEPTION_HUB_ALREADY_ANNOUNCED = 0xe00;

	// Other constants
	const OBJECT_LIST_SEPARATOR = ',';

	/**
	 * Universal node locator of bootstrap node
	 */
	private $bootUnl = '';

	/**
	 * Whether this node is anncounced (keep on FALSE!)
	 * @deprecated
	 */
	private $hubIsAnnounced = FALSE;

	/**
	 * Whether this hub is active (default: FALSE)
	 */
	private $isActive = FALSE;

	/**
	 * Whether this node accepts announcements (default: FALSE)
	 */
	private $acceptAnnouncements = FALSE;

	/**
	 * Whether this node accepts DHT bootstrap requests (default: FALSE)
	 */
	private $acceptDhtBootstrap = FALSE;

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct ($className) {
		// Call parent constructor
		parent::__construct($className);

		// Get a wrapper instance
		$wrapperInstance = DatabaseWrapperFactory::createWrapperByConfiguredName('node_info_db_wrapper_class');

		// Set it here
		$this->setWrapperInstance($wrapperInstance);

		// Get a crypto instance
		$cryptoInstance = ObjectFactory::createObjectByConfiguredName('crypto_class');

		// Set it here
		$this->setCryptoInstance($cryptoInstance);

		// Set the node instance in registry
		Registry::getRegistry()->addInstance('node', $this);

		// Init state which sets the state to 'init'
		$this->initState();
	}

	/**
	 * Initializes the node's state which sets it to 'init'
	 *
	 * @return	void
	 */
	private function initState() {
		// Get the state factory and create the initial state.
		NodeStateFactory::createNodeStateInstanceByName('init');
	}

	/**
	 * Generates a private key and hashes it (for speeding up things)
	 *
	 * @param	$searchInstance		An instance of a LocalSearchCriteria class
	 * @return	void
	 */
	private function generatePrivateKeyAndHash (LocalSearchCriteria $searchInstance) {
		// Get an RNG instance
		$rngInstance = ObjectFactory::createObjectByConfiguredName('rng_class');

		// Generate a pseudo-random string
		$randomString = $rngInstance->randomString(255);

		// Hash and encrypt the string so we become a node id (also documented as "hub id")
		$this->setPrivateKey($this->getCryptoInstance()->encryptString($randomString));
		$this->setPrivateKeyHash($this->getCryptoInstance()->hashString($this->getPrivateKey()));

		// Register the node id with our wrapper
		$this->getWrapperInstance()->registerPrivateKey($this, $this->getRequestInstance(), $searchInstance);

		// Output message
		self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: Created new private key with hash: ' . $this->getPrivateKeyHash() . '');
	}

	/**
	 * Generates a random string from various data inluding UUID if PECL
	 * extension uuid is installed.
	 *
	 * @param	$length			Length of the random part
	 * @return	$randomString	Random string
	 * @todo	Make this code more generic and move it to CryptoHelper or
	 */
	protected function generateRamdomString ($length) {
		// Get an RNG instance
		$rngInstance = ObjectFactory::createObjectByConfiguredName('rng_class');

		// Generate a pseudo-random string
		$randomString = $rngInstance->randomString($length) . ':' . $this->getBootUniversalNodeLocator() . ':' . $this->getRequestInstance()->getRequestElement('mode');

		// Add UUID for even more entropy for the hasher
		$randomString .= $this->getCryptoInstance()->createUuid();

		// Return it
		return $randomString;
	}

	/**
	 * Getter for boot UNL (Universal Node Locator)
	 *
	 * @return	$bootUnl	The UNL (Universal Node Locator) of the boot node
	 */
	protected final function getBootUniversalNodeLocator () {
		return $this->bootUnl;
	}

	/**
	 * Checks whether the given IP address matches one of the bootstrap nodes
	 *
	 * @param	$remoteAddr		IP address to checkout against our bootstrapping list
	 * @return	$isFound		Whether the IP is found
	 */
	protected function ifAddressMatchesBootstrapNodes ($remoteAddr) {
		// By default nothing is found
		$isFound = FALSE;

		// Run through all configured IPs
		foreach (explode(BaseHubSystem::BOOTSTRAP_NODES_SEPARATOR, $this->getConfigInstance()->getConfigEntry('hub_bootstrap_nodes')) as $unl) {
			// Does it match?
			if ($unl == $remoteAddr) {
				// Found it!
				$isFound = TRUE;

				// Remember the UNL
				$this->bootUnl = $unl;

				// Output message
				self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: ' . __FUNCTION__ . '[' . __METHOD__ . ':' . __LINE__ . ']: UNL matches remote address ' . $unl . '.');

				// Stop further searching
				break;
			} elseif ($unl == $this->getConfigInstance()->getConfigEntry('node_listen_addr')) {
				/*
				 * IP matches listen address. At this point we really don't care
				 * if we can really listen on that address
				 */
				$isFound = TRUE;

				// Remember the port number
				$this->bootUnl = $unl;

				// Output message
				self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: ' . __FUNCTION__ . '[' . __METHOD__ . ':' . __LINE__ . ']: UNL matches listen address ' . $unl . '.');

				// Stop further searching
				break;
			}
		} // END - foreach

		// Return the result
		return $isFound;
	}

	/**
	 * Tries to detect own UNL (Universal Node Locator)
	 *
	 * @return	$unl	Node's own universal node locator
	 */
	public function detectOwnUniversalNodeLocator () {
		// Is "cache" set?
		if (!isset($GLOBALS[__METHOD__])) {
			// Get the UNL array back
			$unlData = $this->getUniversalNodeLocatorArray();

			// There are 2 UNLs, internal and external.
			if ($this->getConfigInstance()->getConfigEntry('allow_publish_internal_address') == 'N') {
				// Public "external" UNL address
				$GLOBALS[__METHOD__] = $unlData[NodeInformationDatabaseWrapper::DB_COLUMN_EXTERNAL_UNL];
			} else {
				// Non-public "internal" UNL address
				$GLOBALS[__METHOD__] = $unlData[NodeInformationDatabaseWrapper::DB_COLUMN_INTERNAL_UNL];
			}
		} // END - if

		// Return it
		return $GLOBALS[__METHOD__];
	}

	/**
	 * Outputs the console teaser. This should only be executed on startup or
	 * full restarts. This method generates some space around the teaser.
	 *
	 * @return	void
	 */
	public function outputConsoleTeaser () {
		// Get the app instance (for shortening our code)
		$app = $this->getApplicationInstance();

		// Output all lines
		self::createDebugInstance(__CLASS__)->debugOutput(' ');
		self::createDebugInstance(__CLASS__)->debugOutput($app->getAppName() . ' v' . $app->getAppVersion() . ' - ' . $this->getRequestInstance()->getRequestElement('mode') . ' mode active');
		self::createDebugInstance(__CLASS__)->debugOutput('Copyright (c) 2007 - 2008 Roland Haeder, 2009 - 2015 Hub Developer Team');
		self::createDebugInstance(__CLASS__)->debugOutput(' ');
		self::createDebugInstance(__CLASS__)->debugOutput('This program comes with ABSOLUTELY NO WARRANTY; for details see docs/COPYING.');
		self::createDebugInstance(__CLASS__)->debugOutput('This is free software, and you are welcome to redistribute it under certain');
		self::createDebugInstance(__CLASS__)->debugOutput('conditions; see docs/COPYING for details.');
		self::createDebugInstance(__CLASS__)->debugOutput(' ');
	}

	/**
	 * Generic method to acquire a hub-id. On first run this generates a new one
	 * based on many pseudo-random data. On any later run, unless the id
	 * got not removed from database, it will be restored from the database.
	 *
	 * @param	$requestInstance	A Requestable class
	 * @param	$responseInstance	A Responseable class
	 * @return	void
	 */
	public function bootstrapAcquireNodeId (Requestable $requestInstance, Responseable $responseInstance) {
		// Is there a node id?
		if ($this->getWrapperInstance()->ifNodeDataIsFound($this)) {
			// Get the node id from result and set it
			$this->setNodeId($this->getField(NodeInformationDatabaseWrapper::DB_COLUMN_NODE_ID));

			// Output message
			self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: Re-using found node-id: ' . $this->getNodeId() . '');
		} else {
			// Get an RNG instance
			$rngInstance = ObjectFactory::createObjectByConfiguredName('rng_class');

			// Generate a pseudo-random string
			$randomString = $rngInstance->randomString(255);

			// Hash and encrypt the string so we become a node id (also documented as "hub id")
			$this->setNodeId($this->getCryptoInstance()->hashString($this->getCryptoInstance()->encryptString($randomString)));

			// Register the node id with our wrapper
			$this->getWrapperInstance()->registerNodeId($this, $this->getRequestInstance());

			// Output message
			self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: Created new node-id: ' . $this->getNodeId() . '');
		}
	}

	/**
	 * Generates a session id which will be sent to the other hubs and peers
	 *
	 * @return	void
	 */
	public function bootstrapGenerateSessionId () {
		// Now get a search criteria instance
		$searchInstance = ObjectFactory::createObjectByConfiguredName('search_criteria_class');

		// Search for the node number one which is hard-coded the default
		$searchInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_NODE_NR  , 1);
		$searchInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_NODE_MODE, $this->getRequestInstance()->getRequestElement('mode'));
		$searchInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_NODE_ID  , $this->getNodeId());
		$searchInstance->setLimit(1);

		// Remember it for later usage
		$this->setSearchInstance($searchInstance);

		// Get a random string
		$randomString = $this->generateRamdomString(255);

		// Hash and encrypt the string so we become a "node id" aka Hub-Id
		$this->setSessionId($this->getCryptoInstance()->hashString($this->getCryptoInstance()->encryptString($randomString)));

		// Register the node id with our wrapper
		$this->getWrapperInstance()->registerSessionId($this, $this->getRequestInstance(), $searchInstance);

		// Output message
		self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: Created new session-id: ' . $this->getSessionId() . '');

		// Change the state because the node has auired a hub id
		$this->getStateInstance()->nodeGeneratedSessionId();
	}

	/**
	 * Generate a private key for en-/decryption
	 *
	 * @return	void
	 */
	public function bootstrapGeneratePrivateKey () {
		// Is it valid?
		if ($this->getWrapperInstance()->ifNodeDataIsFound($this)) {
			// Is the element set?
			if (is_null($this->getField(NodeInformationDatabaseWrapper::DB_COLUMN_PRIVATE_KEY))) {
				/*
				 * Auto-generate the private key for e.g. out-dated database
				 * "tables". This allows a smooth update for the underlaying
				 * database table.
				 */
				$this->generatePrivateKeyAndHash($this->getSearchInstance());
			} else {
				// Get the node id from result and set it
				$this->setPrivateKey(base64_decode($this->getField(NodeInformationDatabaseWrapper::DB_COLUMN_PRIVATE_KEY)));
				$this->setPrivateKeyHash($this->getField(NodeInformationDatabaseWrapper::DB_COLUMN_PRIVATE_KEY_HASH));

				// Output message
				self::createDebugInstance(__CLASS__)->debugOutput('BOOTSTRAP: Re-using found private key hash: ' . $this->getPrivateKeyHash() . '');
			}
		} else {
			/*
			 * Generate it in a private method (no confusion with 'private
			 * method access' and 'private key' here! ;-)).
			 */
			$this->generatePrivateKeyAndHash($this->getSearchInstance());
		}
	}

	/**
	 * Adds hub data elements to a given dataset instance
	 *
	 * @param	$criteriaInstance	An instance of a storeable criteria
	 * @param	$requestInstance	An instance of a Requestable class
	 * @return	void
	 */
	public function addElementsToDataSet (StoreableCriteria $criteriaInstance, Requestable $requestInstance = NULL) {
		// Make sure the request instance is set as it is not optional.
		assert($requestInstance instanceof Requestable);

		// Add node number and type
		$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_NODE_NR  , 1);
		$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_NODE_MODE, $requestInstance->getRequestElement('mode'));

		// Add the node id
		$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_NODE_ID, $this->getNodeId());

		// Add the session id if acquired
		if ($this->getSessionId() != '') {
			$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_SESSION_ID, $this->getSessionId());
		} // END - if

		// Add the private key if acquired
		if ($this->getPrivateKey() != '') {
			$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_PRIVATE_KEY, base64_encode($this->getPrivateKey()));
			$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_PRIVATE_KEY_HASH, $this->getPrivateKeyHash());
		} // END - if

		// Add own external and internal addresses as UNLs
		$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_INTERNAL_UNL, HubTools::determineOwnInternalAddress());
		$criteriaInstance->addCriteria(NodeInformationDatabaseWrapper::DB_COLUMN_EXTERNAL_UNL, HubTools::determineOwnExternalAddress());
	}

	/**
	 * Updates a given field with new value
	 *
	 * @param	$fieldName		Field to update
	 * @param	$fieldValue		New value to store
	 * @return	void
	 * @throws	DatabaseUpdateSupportException	If this class does not support database updates
	 * @todo	Try to make this method more generic so we can move it in BaseFrameworkSystem
	 */
	public function updateDatabaseField ($fieldName, $fieldValue) {
		// Unfinished
		$this->partialStub('Unfinished!');
		return;

		// Get a critieria instance
		$searchInstance = ObjectFactory::createObjectByConfiguredName('search_criteria_class');

		// Add search criteria
		$searchInstance->addCriteria(UserDatabaseWrapper::DB_COLUMN_USERNAME, $this->getUserName());
		$searchInstance->setLimit(1);

		// Now get another criteria
		$updateInstance = ObjectFactory::createObjectByConfiguredName('update_criteria_class');

		// Add criteria entry which we shall update
		$updateInstance->addCriteria($fieldName, $fieldValue);

		// Add the search criteria for searching for the right entry
		$updateInstance->setSearchInstance($searchInstance);

		// Set wrapper class name
		$updateInstance->setWrapperConfigEntry('user_db_wrapper_class');

		// Remember the update in database result
		$this->getResultInstance()->add2UpdateQueue($updateInstance);
	}

	/**
	 * Announces this hub to the upper (bootstrap or list) hubs. After this is
	 * successfully done the given task is unregistered from the handler. This
	 * might look a bit overloaded here but the announcement phase isn't a
	 * simple "Hello there" message, it may later on also contain more
	 * informations like the object list.
	 *
	 * @param	$taskInstance	The task instance running this announcement
	 * @return	void
	 * @throws	NodeAlreadyAnnouncedException	If this hub is already announced
	 * @todo	Change the first if() block to check for a specific state
	 */
	public function announceToUpperNodes (Taskable $taskInstance) {
		// Is this hub node announced?
		if ($this->hubIsAnnounced === TRUE) {
			// Already announced!
			throw new NodeAlreadyAnnouncedException($this, self::EXCEPTION_HUB_ALREADY_ANNOUNCED);
		} // END - if

		// Debug output
		self::createDebugInstance(__CLASS__)->debugOutput('HUB-Announcement: START (taskInstance=' . $taskInstance->__toString(). ')');

		// Get a helper instance
		$helperInstance = ObjectFactory::createObjectByConfiguredName('node_announcement_helper_class');

		// Load the announcement descriptor
		$helperInstance->loadDescriptorXml($this);

		// Compile all variables
		$helperInstance->getTemplateInstance()->compileConfigInVariables();

		// "Publish" the descriptor by sending it to the bootstrap/list nodes
		$helperInstance->sendPackage($this);

		// Change the state, this should be the last line except debug output
		$this->getStateInstance()->nodeAnnouncingToUpperHubs();

		// Debug output
		self::createDebugInstance(__CLASS__)->debugOutput('HUB-Announcement: FINISHED');
	}

	/**
	 * Does a self-connect attempt on the public IP address. This should make
	 * it sure, we are reachable from outside world. For this kind of package we
	 * don't need that overload we have in the announcement phase.
	 *
	 * @param	$taskInstance	The task instance running this announcement
	 * @return	void
	 */
	public function doSelfConnection (Taskable $taskInstance) {
		// Debug output
		self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: Self Connection: START (taskInstance=' . $taskInstance->__toString(). ')');

		// Get a helper instance
		$helperInstance = ObjectFactory::createObjectByConfiguredName('node_self_connect_helper_class', array($this));

		// Load the descriptor (XML) file
		$helperInstance->loadDescriptorXml($this);

		// Compile all variables
		$helperInstance->getTemplateInstance()->compileConfigInVariables();

		// And send the package away
		$helperInstance->sendPackage($this);

		// Debug output
		self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: Self Connection: FINISHED');
	}

	/**
	 * Activates the hub by doing some final preparation and setting
	 * $hubIsActive to TRUE.
	 *
	 * @param	$requestInstance	A Requestable class
	 * @param	$responseInstance	A Responseable class
	 * @return	void
	 */
	public function activateNode (Requestable $requestInstance, Responseable $responseInstance) {
		// Checks whether a listener is still active and shuts it down if one
		// is still listening.
		if (($this->determineIfListenerIsActive()) && ($this->isNodeActive())) {
			// Shutdown them down before they can hurt anything
			$this->shutdownListenerPool();
		} // END - if

		// Get the controller here
		$controllerInstance = Registry::getRegistry()->getInstance('controller');

		// Run all filters for the hub activation
		$controllerInstance->executeActivationFilters($requestInstance, $responseInstance);

		// ----------------------- Last step from here ------------------------
		// Activate the hub. This is ALWAYS the last step in this method
		$this->getStateInstance()->nodeIsActivated();
		// ---------------------- Last step until here ------------------------
	}

	/**
	 * Initializes the listener pool (class)
	 *
	 * @return	void
	 */
	public function initializeListenerPool () {
		// Debug output
		self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: Initialize listener: START');

		// Get a new pool instance
		$this->setListenerPoolInstance(ObjectFactory::createObjectByConfiguredName('listener_pool_class', array($this)));

		// Get an instance of the low-level listener
		$listenerInstance = ObjectFactory::createObjectByConfiguredName('tcp_listener_class', array($this));

		// Setup address and port
		$listenerInstance->setListenAddressByConfiguration('node_listen_addr');

		/*
		 * All nodes can now use the same configuration entry because it can be
		 * customized in config-local.php.
		 */
		$listenerInstance->setListenPortByConfiguration('node_listen_port');

		// Initialize the listener
		$listenerInstance->initListener();

		// Get a decorator class
		$decoratorInstance = ObjectFactory::createObjectByConfiguredName('node_tcp_listener_class', array($listenerInstance));

		// Add this listener to the pool
		$this->getListenerPoolInstance()->addListener($decoratorInstance);

		// Get a decorator class
		$decoratorInstance = ObjectFactory::createObjectByConfiguredName('client_tcp_listener_class', array($listenerInstance));

		// Add this listener to the pool
		$this->getListenerPoolInstance()->addListener($decoratorInstance);

		// Get an instance of the low-level listener
		$listenerInstance = ObjectFactory::createObjectByConfiguredName('udp_listener_class', array($this));

		// Setup address and port
		$listenerInstance->setListenAddressByConfiguration('node_listen_addr');

		/*
		 * All nodes can now use the same configuration entry because it can be
		 * customized in config-local.php.
		 */
		$listenerInstance->setListenPortByConfiguration('node_listen_port');

		// Initialize the listener
		$listenerInstance->initListener();

		// Get a decorator class
		$decoratorInstance = ObjectFactory::createObjectByConfiguredName('node_udp_listener_class', array($listenerInstance));

		// Add this listener to the pool
		$this->getListenerPoolInstance()->addListener($decoratorInstance);

		// Get a decorator class
		$decoratorInstance = ObjectFactory::createObjectByConfiguredName('client_udp_listener_class', array($listenerInstance));

		// Add this listener to the pool
		$this->getListenerPoolInstance()->addListener($decoratorInstance);

		// Debug output
		self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: Initialize listener: FINISHED.');
	}

	/**
	 * Getter for isActive attribute
	 *
	 * @return	$isActive	Whether the hub is active
	 */
	public final function isNodeActive () {
		return $this->isActive;
	}

	/**
	 * Enables (default) or disables isActive flag
	 *
	 * @param	$isActive	Whether the hub is active
	 * @return	void
	 */
	public final function enableIsActive ($isActive = TRUE) {
		$this->isActive = (bool) $isActive;
	}

	/**
	 * Checks whether this node accepts announcements
	 *
	 * @return	$acceptAnnouncements	Whether this node accepts announcements
	 */
	public final function isAcceptingAnnouncements () {
		// Check it (this node must be active and not shutdown!)
		$acceptAnnouncements = (($this->acceptAnnouncements === TRUE) && ($this->isNodeActive()));

		// Return it
		return $acceptAnnouncements;
	}

	/**
	 * Checks whether this node accepts DHT bootstrap requests
	 *
	 * @return	$acceptDhtBootstrap	Whether this node accepts DHT bootstrap requests
	 */
	public final function isAcceptingDhtBootstrap () {
		// Check it (this node must be active and not shutdown!)
		$acceptDhtBootstrap = (($this->acceptDhtBootstrap === TRUE) && ($this->isNodeActive()));

		// Return it
		return $acceptDhtBootstrap;
	}

	/**
	 * Checks whether this node has attempted to announce itself
	 *
	 * @return	$hasAnnounced	Whether this node has attempted to announce itself
	 * @todo	Add checking if this node has been announced to the sender node
	 */
	public function ifNodeIsAnnouncing () {
		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: ifNodeIsAnnouncing(): state=' . $this->getStateInstance()->getStateName());

		// Simply check the state of this node
		$hasAnnounced = ($this->getStateInstance() instanceof NodeAnnouncingState);

		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: ifNodeIsAnnouncing(): hasAnnounced=' . intval($hasAnnounced));

		// Return it
		return $hasAnnounced;
	}

	/**
	 * Checks whether this node has attempted to announce itself and completed it
	 *
	 * @return	$hasAnnouncementCompleted	Whether this node has attempted to announce itself and completed it
	 * @todo	Add checking if this node has been announced to the sender node
	 */
	public function ifNodeHasAnnouncementCompleted () {
		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: ifNodeHasAnnouncementCompleted(): state=' . $this->getStateInstance()->getStateName());

		// Simply check the state of this node
		$hasAnnouncementCompleted = ($this->getStateInstance() instanceof NodeAnnouncementCompletedState);

		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: ifNodeHasAnnouncementCompleted(): hasAnnouncementCompleted=' . intval($hasAnnouncementCompleted));

		// Return it
		return $hasAnnouncementCompleted;
	}

	/**
	 * Enables whether this node accepts announcements
	 *
	 * @param	$acceptAnnouncements	Whether this node accepts announcements (default: TRUE)
	 * @return	void
	 */
	protected final function enableAcceptingAnnouncements ($acceptAnnouncements = TRUE) {
		$this->acceptAnnouncements = $acceptAnnouncements;
	}

	/**
	 * Enables whether this node accepts DHT bootstrap requests
	 *
	 * @param	$acceptDhtBootstrap	Whether this node accepts DHT bootstrap requests (default: TRUE)
	 * @return	void
	 */
	public final function enableAcceptDhtBootstrap ($acceptDhtBootstrap = TRUE) {
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: Enabling DHT bootstrap requests ...');
		$this->acceptDhtBootstrap = $acceptDhtBootstrap;
	}

	/**
	 * Checks wether this node is accepting node-list requests
	 *
	 * @return	$acceptsRequest		Wether this node accepts node-list requests
	 */
	public function isAcceptingNodeListRequests () {
		/*
		 * Only 'regular' nodes does not accept such requests, checking
		 * HubRegularNode is faster, but if e.g. HubRegularI2PNode will be
		 * added then the next check will be TRUE.
		 */
		$acceptsRequest = ((!$this instanceof HubRegularNode) && ($this->getRequestInstance()->getRequestElement('mode') != self::NODE_TYPE_REGULAR));

		// Return it
		return $acceptsRequest;
	}

	/**
	 * Determines an instance of a LocateableNode class
	 *
	 * @return	$unlInstance	An instance of a LocateableNode class for this node
	 */
	public function determineUniversalNodeLocator () {
		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: CALLED!');

		// Determine UNL based on this node:
		// 1) Get discovery class
		$discoveryInstance = ObjectFactory::createObjectByConfiguredName('unl_discovery_class');

		// 2) "Determine" it
		$unlInstance = $discoveryInstance->discoverUniversalNodeLocatorByNode($this);

		// 3) Return it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: unlInstance= ' . $unlInstance->__toString() . ' - EXIT!');
		return $unlInstance;
	}

	/**
	 * "Getter" for an array of an instance of a LocateableNode class
	 *
	 * @return	$unlData	An array from an instance of a LocateableNode class for this node
	 */
	public final function getUniversalNodeLocatorArray () {
		// Debug message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: CALLED!');

		// Get the Universal Node Locator (UNL) instance
		$unlInstance = $this->determineUniversalNodeLocator();

		// Make sure the instance is valid
		if (!$unlInstance instanceof LocateableNode) {
			// No valid instance, so better debug this
			$this->debugBackTrace('unlInstance[' . gettype($unlInstance) . ']=' . $unlInstance);
		} // END - if

		// ... and the array from it
		$unlData = $unlInstance->getUnlData();

		// Return it
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: EXIT!');
		return $unlData;
	}

	/**
	 * Updates/refreshes node data (e.g. status).
	 *
	 * @return	void
	 * @todo	Find more to do here
	 */
	public function updateNodeData () {
		// Set some dummy configuration entries, e.g. node_status
		$this->getConfigInstance()->setConfigEntry('node_status', $this->getStateInstance()->getStateName());
	}

	/**
	 * Handles message answer by given data array
	 *
	 * @param	$messageData		A valid answer message data array
	 * @param	$packageInstance	An instance of a Receivable class
	 * @return	void
	 * @todo	Handle thrown exception
	 */
	public function handleAnswerStatusByMessageData (array $messageData, Receivable $packageInstance) {
		// Is it not empty?
		assert(!empty($messageData[BaseXmlAnswerTemplateEngine::ANSWER_STATUS]));

		// Construct configuration entry for handling class' name
		$classConfigEntry = strtolower($messageData[NetworkPackage::MESSAGE_ARRAY_TYPE] . '_status_' . $messageData[BaseXmlAnswerTemplateEngine::ANSWER_STATUS]) . '_handler_class';

		// Try to get a class
		$handlerInstance = ObjectFactory::createObjectByConfiguredName($classConfigEntry);

		// Handle it there
		$handlerInstance->handleAnswerMessageData($messageData, $packageInstance);
	}

	/**
	 * "Getter" for an array of all accepted object types
	 *
	 * @return	$objectList		Array of all accepted object types
	 */
	public function getListFromAcceptedObjectTypes () {
		// Get registry instance
		$objectRegistryInstance = ObjectTypeRegistryFactory::createObjectTypeRegistryInstance();

		// Get all entries
		$objectList = $objectRegistryInstance->getEntries(XmlObjectRegistryTemplateEngine::OBJECT_TYPE_DATA_NAME);

		// ... and return it
		return $objectList;
	}

	/**
	 * Adds all required elements from given array into data set instance
	 *
	 * @param	$dataSetInstance	An instance of a StoreableCriteria class
	 * @param	$nodeData			An array with valid node data
	 * @return	void
	 */
	public function addArrayToDataSet (StoreableCriteria $dataSetInstance, array $nodeData) {
		// Add all data the array provides
		foreach (NodeDistributedHashTableDatabaseWrapper::getAllElements() as $element) {
			// Is the element there?
			if (isset($nodeData[$element])) {
				// Add it
				$dataSetInstance->addCriteria($element, $nodeData[$element]);
			} else {
				// Output warning message
				/* DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('NODE[' . __METHOD__ . ':' . __LINE__ . ']: addArrayToDataSet(): Element ' . $element . ' not found in nodeData array.');
			}
		} // END - foreac
	}
}

// [EOF]
?>
