<?php
/**
 * An Bootstrap template engine class for XML templates
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2015 Hub Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 * @todo		This template engine does not make use of setTemplateType()
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class XmlDhtBootstrapTemplateEngine extends BaseXmlTemplateEngine implements CompileableTemplate, Registerable {
	/**
	 * Some XML nodes must be available for later data extraction
	 */
	const DHT_BOOTSTRAP_DATA_BOOTSTRAP_DATA   = 'dht-bootstrap-data';
	const DHT_BOOTSTRAP_DATA_SESSION_ID       = 'session-id';
	const DHT_BOOTSTRAP_DATA_NODE_STATUS      = 'node-status';
	const DHT_BOOTSTRAP_DATA_NODE_MODE        = 'node-mode';
	const DHT_BOOTSTRAP_DATA_EXTERNAL_ADDRESS = 'external-address';
	const DHT_BOOTSTRAP_DATA_INTERNAL_ADDRESS = 'internal-address';

	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	protected function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);

		// Init array
		$this->setSubNodes(array(
			self::DHT_BOOTSTRAP_DATA_BOOTSTRAP_DATA,
			'listener',
			self::DHT_BOOTSTRAP_DATA_NODE_STATUS,
			self::DHT_BOOTSTRAP_DATA_NODE_MODE,
			self::DHT_BOOTSTRAP_DATA_SESSION_ID,
			self::DHT_BOOTSTRAP_DATA_EXTERNAL_ADDRESS,
			self::DHT_BOOTSTRAP_DATA_INTERNAL_ADDRESS
		));
	}

	/**
	 * Creates an instance of the class TemplateEngine and prepares it for usage
	 *
	 * @return	$templateInstance		An instance of TemplateEngine
	 * @throws	BasePathIsEmptyException		If the provided $templateBasePath is empty
	 * @throws	InvalidBasePathStringException	If $templateBasePath is no string
	 * @throws	BasePathIsNoDirectoryException	If $templateBasePath is no
	 *											directory or not found
	 * @throws	BasePathReadProtectedException	If $templateBasePath is
	 *											read-protected
	 */
	public static final function createXmlDhtBootstrapTemplateEngine () {
		// Get a new instance
		$templateInstance = new XmlDhtBootstrapTemplateEngine();

		// Init template instance
		$templateInstance->initXmlTemplateEngine('node', 'dht_bootstrap');

		// Return the prepared instance
		return $templateInstance;
	}

	/**
	 * Currently not used
	 *
	 * @param	$resource		XML parser resource (currently ignored)
	 * @param	$characters		Characters to handle
	 * @return	void
	 */
	public function characterHandler ($resource, $characters) {
		// Trim all spaces away
		$characters = trim($characters);

		// Is this string empty?
		if (empty($characters)) {
			// Then skip it silently
			return;
		} // END - if

		/*
		 * Assign the found characters to variable and use the last entry from
		 * stack as the name.
		 */
		parent::assignVariable($this->getStackInstance()->getNamed('dht_bootstrap'), $characters);
	}

	/**
	 * Getter for cache file (FQFN)
	 *
	 * @return	$fqfn	Full-qualified file name of the menu cache
	 */
	public function getMenuCacheFqfn () {
		$this->partialStub('Please implement this method.');
	}

	/**
	 * Starts the dht bootstrap
	 *
	 * @return	void
	 */
	protected function startDhtBootstrap () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', 'dht-bootstrap');
	}

	/**
	 * Starts the dht bootstrap data
	 *
	 * @return	void
	 */
	protected function startDhtBootstrapData () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', self::DHT_BOOTSTRAP_DATA_BOOTSTRAP_DATA);
	}

	/**
	 * Starts the node status
	 *
	 * @return	void
	 */
	protected function startNodeStatus () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', self::DHT_BOOTSTRAP_DATA_NODE_STATUS);
	}

	/**
	 * Starts the node-mode
	 *
	 * @return	void
	 */
	protected function startNodeMode () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', self::DHT_BOOTSTRAP_DATA_NODE_MODE);
	}

	/**
	 * Starts the listener
	 *
	 * @return	void
	 */
	protected function startListener () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', 'listener');
	}

	/**
	 * Starts the public ip
	 *
	 * @return	void
	 */
	protected function startExternalAddress () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', self::DHT_BOOTSTRAP_DATA_EXTERNAL_ADDRESS);
	}

	/**
	 * Starts the private ip
	 *
	 * @return	void
	 */
	protected function startInternalAddress () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', self::DHT_BOOTSTRAP_DATA_INTERNAL_ADDRESS);
	}

	/**
	 * Starts the session id
	 *
	 * @return	void
	 */
	protected function startSessionId () {
		// Push the node name on the stacker
		$this->getStackInstance()->pushNamed('dht_bootstrap', self::DHT_BOOTSTRAP_DATA_SESSION_ID);
	}

	/**
	 * Finishes the session id
	 *
	 * @return	void
	 */
	protected function finishSessionId () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}

	/**
	 * Finishes the private ip
	 *
	 * @return	void
	 */
	protected function finishInternalAddress () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}

	/**
	 * Finishes the public ip
	 *
	 * @return	void
	 */
	protected function finishExternalAddress () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}

	/**
	 * Finishes the listener
	 *
	 * @return	void
	 */
	protected function finishListener () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}

	/**
	 * Finishes the node mode
	 *
	 * @return	void
	 */
	protected function finishNodeMode () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}

	/**
	 * Finishes the node status
	 *
	 * @return	void
	 */
	protected function finishNodeStatus () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}

	/**
	 * Finishes the dht bootstrap data
	 *
	 * @return	void
	 */
	protected function finishDhtBootstrapData () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}

	/**
	 * Finishes the dht bootstrap
	 *
	 * @return	void
	 */
	protected function finishDhtBootstrap () {
		// Pop the last entry
		$this->getStackInstance()->popNamed('dht_bootstrap');
	}
}

// [EOF]
?>
